@@ -3,12 +3,12 @@
  * fe-exec.c
  *	  functions related to sending a query down to the backend
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/interfaces/libpq/fe-exec.c,v 1.113.2.1 2002/04/08 06:21:31 ishii Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-exec.c,v 1.122.2.1 2003/11/30 20:52:37 joe Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -115,7 +115,8 @@ PQescapeString(char *to, const char *from, size_t length)
  *		'\0' == ASCII  0 == \\000
  *		'\'' == ASCII 39 == \'
  *		'\\' == ASCII 92 == \\\\
- *		anything >= 0x80 ---> \\ooo (where ooo is an octal expression)
+ *		anything < 0x20, or > 0x7e ---> \\ooo
+ *                                      (where ooo is an octal expression)
  */
 unsigned char *
 PQescapeBytea(unsigned char *bintext, size_t binlen, size_t *bytealen)
@@ -134,8 +135,8 @@ PQescapeBytea(unsigned char *bintext, size_t binlen, size_t *bytealen)
 	vp = bintext;
 	for (i = binlen; i > 0; i--, vp++)
 	{
-		if (*vp == 0 || *vp >= 0x80)
-			len += 5;	/* '5' is for '\\ooo' */
+		if (*vp < 0x20 || *vp > 0x7e)
+			len += 5;			/* '5' is for '\\ooo' */
 		else if (*vp == '\'')
 			len += 2;
 		else if (*vp == '\\')
@@ -153,9 +154,9 @@ PQescapeBytea(unsigned char *bintext, size_t binlen, size_t *bytealen)
 
 	for (i = binlen; i > 0; i--, vp++)
 	{
-		if (*vp == 0 || *vp >= 0x80)
+		if (*vp < 0x20 || *vp > 0x7e)
 		{
-			(void)sprintf(rp,"\\\\%03o",*vp);
+			(void) sprintf(rp, "\\\\%03o", *vp);
 			rp += 5;
 		}
 		else if (*vp == '\'')
@@ -180,6 +181,110 @@ PQescapeBytea(unsigned char *bintext, size_t binlen, size_t *bytealen)
 	return result;
 }
 
+/*
+ *		PQunescapeBytea - converts the null terminated string representation
+ *		of a bytea, strtext, into binary, filling a buffer. It returns a
+ *		pointer to the buffer which is NULL on error, and the size of the
+ *		buffer in retbuflen. The pointer may subsequently be used as an
+ *		argument to the function free(3). It is the reverse of PQescapeBytea.
+ *
+ *		The following transformations are reversed:
+ *		'\0' == ASCII  0 == \000
+ *		'\'' == ASCII 39 == \'
+ *		'\\' == ASCII 92 == \\
+ *
+ *		States:
+ *		0	normal		0->1->2->3->4
+ *		1	\			   1->5
+ *		2	\0			   1->6
+ *		3	\00
+ *		4	\000
+ *		5	\'
+ *		6	\\
+ */
+unsigned char *
+PQunescapeBytea(unsigned char *strtext, size_t *retbuflen)
+{
+	size_t		buflen;
+	unsigned char *buffer,
+			   *sp,
+			   *bp;
+	unsigned int state = 0;
+
+	if (strtext == NULL)
+		return NULL;
+	buflen = strlen(strtext);	/* will shrink, also we discover if
+								 * strtext */
+	buffer = (unsigned char *) malloc(buflen);	/* isn't NULL terminated */
+	if (buffer == NULL)
+		return NULL;
+	for (bp = buffer, sp = strtext; *sp != '\0'; bp++, sp++)
+	{
+		switch (state)
+		{
+			case 0:
+				if (*sp == '\\')
+					state = 1;
+				*bp = *sp;
+				break;
+			case 1:
+				if (*sp == '\'')	/* state=5 */
+				{				/* replace \' with 39 */
+					bp--;
+					*bp = '\'';
+					buflen--;
+					state = 0;
+				}
+				else if (*sp == '\\')	/* state=6 */
+				{				/* replace \\ with 92 */
+					bp--;
+					*bp = '\\';
+					buflen--;
+					state = 0;
+				}
+				else
+				{
+					if (isdigit(*sp))
+						state = 2;
+					else
+						state = 0;
+					*bp = *sp;
+				}
+				break;
+			case 2:
+				if (isdigit(*sp))
+					state = 3;
+				else
+					state = 0;
+				*bp = *sp;
+				break;
+			case 3:
+				if (isdigit(*sp))		/* state=4 */
+				{
+					int			v;
+
+					bp -= 3;
+					sscanf(sp - 2, "%03o", &v);
+					*bp = v;
+					buflen -= 3;
+					state = 0;
+				}
+				else
+				{
+					*bp = *sp;
+					state = 0;
+				}
+				break;
+		}
+	}
+	buffer = realloc(buffer, buflen);
+	if (buffer == NULL)
+		return NULL;
+
+	*retbuflen = buflen;
+	return buffer;
+}
+
 /* ----------------
  * Space management for PGresult.
  *
@@ -709,7 +814,7 @@ PQsendQuery(PGconn *conn, const char *query)
  * handleSendFailure: try to clean up after failure to send command.
  *
  * Primarily, what we want to accomplish here is to process an async
- * NOTICE message that the backend might have sent just before it died.
+ * WARNING message that the backend might have sent just before it died.
  *
  * NOTE: this routine should only be called in PGASYNC_IDLE state.
  */
@@ -727,7 +832,7 @@ handleSendFailure(PGconn *conn)
 
 	/*
 	 * Parse any available input messages.	Since we are in PGASYNC_IDLE
-	 * state, only NOTICE and NOTIFY messages will be eaten.
+	 * state, only WARNING and NOTIFY messages will be eaten.
 	 */
 	parseInput(conn);
 }
@@ -801,8 +906,8 @@ parseInput(PGconn *conn)
 			return;
 
 		/*
-		 * NOTIFY and NOTICE messages can happen in any state besides COPY
-		 * OUT; always process them right away.
+		 * NOTIFY and WARNING messages can happen in any state besides
+		 * COPY OUT; always process them right away.
 		 *
 		 * Most other messages should only be processed while in BUSY state.
 		 * (In particular, in READY state we hold off further parsing
@@ -1045,14 +1150,10 @@ getRowDescriptions(PGconn *conn)
 
 		/*
 		 * Since pqGetInt treats 2-byte integers as unsigned, we need to
-		 * coerce the special value "-1" to signed form.  (-1 is sent for
-		 * variable-length fields.)  Formerly, libpq effectively did a
-		 * sign-extension on the 2-byte value by storing it in a signed
-		 * short. Now we only coerce the single value 65535 == -1; values
-		 * 32768..65534 are taken as valid field lengths.
+		 * coerce the result to signed form.
 		 */
-		if (typlen == 0xFFFF)
-			typlen = -1;
+		typlen = (int) ((int16) typlen);
+
 		result->attDescs[i].name = pqResultStrdup(result,
 												  conn->workBuffer.data);
 		result->attDescs[i].typid = typid;
@@ -1417,8 +1518,16 @@ getNotify(PGconn *conn)
 		return EOF;
 	if (pqGets(&conn->workBuffer, conn))
 		return EOF;
-	newNotify = (PGnotify *) malloc(sizeof(PGnotify));
-	strncpy(newNotify->relname, conn->workBuffer.data, NAMEDATALEN);
+
+	/*
+	 * Store the relation name right after the PQnotify structure so it
+	 * can all be freed at once.  We don't use NAMEDATALEN because we
+	 * don't want to tie this interface to a specific server name length.
+	 */
+	newNotify = (PGnotify *) malloc(sizeof(PGnotify) +
+									strlen(conn->workBuffer.data) +1);
+	newNotify->relname = (char *) newNotify + sizeof(PGnotify);
+	strcpy(newNotify->relname, conn->workBuffer.data);
 	newNotify->be_pid = be_pid;
 	DLAddTail(conn->notifyList, DLNewElem(newNotify));
 	return 0;
@@ -2340,3 +2449,11 @@ PQflush(PGconn *conn)
 
 	return (pqFlush(conn));
 }
+
+/* try to force data out, really only useful for non-blocking users.
+ * This implementation actually works for non-blocking connections */
+int
+PQsendSome(PGconn *conn)
+{
+	return pqSendSome(conn);
+}
