@@ -3,7 +3,7 @@
  *			  procedural language
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/pl/plpgsql/src/pl_funcs.c,v 1.17.2.1 2002/05/05 17:38:28 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/pl/plpgsql/src/pl_funcs.c,v 1.23 2002/09/05 00:43:07 tgl Exp $
  *
  *	  This software is copyrighted by Jan Wieck - Hamburg.
  *
@@ -35,17 +35,13 @@
  *
  **********************************************************************/
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <unistd.h>
-#include <fcntl.h>
-#include <string.h>
-#include <ctype.h>
-
 #include "plpgsql.h"
 #include "pl.tab.h"
 
+#include <ctype.h>
+
+#include "mb/pg_wchar.h"
+
 
 /* ----------
  * Local variables for the namestack handling
@@ -92,7 +88,8 @@ plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
 	if (needed > ds->alloc)
 	{
 		/* might have to double more than once, if len is large */
-		do {
+		do
+		{
 			ds->alloc *= 2;
 		} while (needed > ds->alloc);
 		ds->value = repalloc(ds->value, ds->alloc);
@@ -152,6 +149,9 @@ plpgsql_ns_push(char *label)
 {
 	PLpgSQL_ns *new;
 
+	if (label == NULL)
+		label = "";
+
 	new = palloc(sizeof(PLpgSQL_ns));
 	memset(new, 0, sizeof(PLpgSQL_ns));
 	new->upper = ns_current;
@@ -192,9 +192,7 @@ plpgsql_ns_additem(int itemtype, int itemno, char *name)
 	PLpgSQL_ns *ns = ns_current;
 	PLpgSQL_nsitem *nse;
 
-	if (name == NULL)
-		name = "";
-	name = plpgsql_tolower(name);
+	Assert(name != NULL);
 
 	if (ns->items_used == ns->items_alloc)
 	{
@@ -269,9 +267,7 @@ plpgsql_ns_lookup(char *name, char *label)
 				return ns->items[i];
 		}
 		if (ns_localmode)
-		{
 			return NULL;		/* name not found in current namespace */
-		}
 	}
 
 	return NULL;
@@ -322,46 +318,111 @@ plpgsql_ns_rename(char *oldname, char *newname)
 
 
 /* ----------
- * plpgsql_tolower			Translate a string to lower case
- *					but honor "" escaping.
+ * plpgsql_convert_ident
+ *
+ * Convert a possibly-qualified identifier to internal form: handle
+ * double quotes, translate to lower case where not inside quotes,
+ * truncate to NAMEDATALEN.
+ *
+ * There may be several identifiers separated by dots and optional
+ * whitespace.	Each one is converted to a separate palloc'd string.
+ * The caller passes the expected number of identifiers, as well as
+ * a char* array to hold them.	It is an error if we find the wrong
+ * number of identifiers (cf grammar processing of fori_varname).
+ *
+ * NOTE: the input string has already been accepted by the flex lexer,
+ * so we don't need a heckuva lot of error checking here.
  * ----------
  */
-char *
-plpgsql_tolower(char *s)
+void
+plpgsql_convert_ident(const char *s, char **output, int numidents)
 {
-	char	   *sstart = s;
-	char	   *ret;
-	char	   *cp;
-
-	ret = palloc(strlen(s) + 1);
-	cp = ret;
+	const char *sstart = s;
+	int			identctr = 0;
 
+	/* Outer loop over identifiers */
 	while (*s)
 	{
+		char	   *curident;
+		char	   *cp;
+		int			i;
+
+		/* Process current identifier */
+		curident = palloc(strlen(s) + 1);		/* surely enough room */
+		cp = curident;
+
 		if (*s == '"')
 		{
+			/* Quoted identifier: copy, collapsing out doubled quotes */
 			s++;
 			while (*s)
 			{
 				if (*s == '"')
-					break;
+				{
+					if (s[1] != '"')
+						break;
+					s++;
+				}
 				*cp++ = *s++;
 			}
-			if (*s != '"')
-				elog(ERROR, "unterminated \" in name %s", sstart);
+			if (*s != '"')		/* should not happen if lexer checked */
+				elog(ERROR, "unterminated \" in name: %s", sstart);
 			s++;
 		}
 		else
 		{
-			if (isupper((unsigned char) *s))
-				*cp++ = tolower((unsigned char) *s++);
-			else
-				*cp++ = *s++;
+			/*
+			 * Normal identifier: downcase, stop at dot or whitespace.
+			 *
+			 * Note that downcasing is locale-sensitive, following SQL99
+			 * rules for identifiers.  We have already decided that the
+			 * item is not a PLPGSQL keyword.
+			 */
+			while (*s && *s != '.' && !isspace((unsigned char) *s))
+			{
+				if (isupper((unsigned char) *s))
+					*cp++ = tolower((unsigned char) *s++);
+				else
+					*cp++ = *s++;
+			}
+		}
+
+		/* Truncate to NAMEDATALEN */
+		*cp = '\0';
+		i = cp - curident;
+
+		if (i >= NAMEDATALEN)
+		{
+			int			len;
+
+			len = pg_mbcliplen(curident, i, NAMEDATALEN - 1);
+			curident[len] = '\0';
+		}
+
+		/* Pass ident to caller */
+		if (identctr < numidents)
+			output[identctr++] = curident;
+		else
+			elog(ERROR, "Qualified identifier cannot be used here: %s",
+				 sstart);
+
+		/* If not done, skip whitespace, dot, whitespace */
+		if (*s)
+		{
+			while (*s && isspace((unsigned char) *s))
+				s++;
+			if (*s++ != '.')
+				elog(ERROR, "Expected dot between identifiers: %s", sstart);
+			while (*s && isspace((unsigned char) *s))
+				s++;
+			if (*s == '\0')
+				elog(ERROR, "Expected another identifier: %s", sstart);
 		}
 	}
-	*cp = '\0';
 
-	return ret;
+	if (identctr != numidents)
+		elog(ERROR, "Improperly qualified identifier: %s",
+			 sstart);
 }
 
 
@@ -393,6 +454,8 @@ plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
 			return "exit";
 		case PLPGSQL_STMT_RETURN:
 			return "return";
+		case PLPGSQL_STMT_RETURN_NEXT:
+			return "return next";
 		case PLPGSQL_STMT_RAISE:
 			return "raise";
 		case PLPGSQL_STMT_EXECSQL:
@@ -432,6 +495,7 @@ static void dump_fors(PLpgSQL_stmt_fors * stmt);
 static void dump_select(PLpgSQL_stmt_select * stmt);
 static void dump_exit(PLpgSQL_stmt_exit * stmt);
 static void dump_return(PLpgSQL_stmt_return * stmt);
+static void dump_return_next(PLpgSQL_stmt_return_next * stmt);
 static void dump_raise(PLpgSQL_stmt_raise * stmt);
 static void dump_execsql(PLpgSQL_stmt_execsql * stmt);
 static void dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt);
@@ -488,6 +552,9 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_RETURN:
 			dump_return((PLpgSQL_stmt_return *) stmt);
 			break;
+		case PLPGSQL_STMT_RETURN_NEXT:
+			dump_return_next((PLpgSQL_stmt_return_next *) stmt);
+			break;
 		case PLPGSQL_STMT_RAISE:
 			dump_raise((PLpgSQL_stmt_raise *) stmt);
 			break;
@@ -771,6 +838,20 @@ dump_return(PLpgSQL_stmt_return * stmt)
 	printf("\n");
 }
 
+static void
+dump_return_next(PLpgSQL_stmt_return_next * stmt)
+{
+	dump_ind();
+	printf("RETURN NEXT ");
+	if (stmt->rec != NULL)
+		printf("target = %d %s\n", stmt->rec->recno, stmt->rec->refname);
+	else if (stmt->row != NULL)
+		printf("target = %d %s\n", stmt->row->rowno, stmt->row->refname);
+	else if (stmt->expr != NULL)
+		dump_expr(stmt->expr);
+	printf("\n");
+}
+
 static void
 dump_raise(PLpgSQL_stmt_raise * stmt)
 {
